from .mcda import MCDA, pd, np, LinearMinMax_

class PROMETHEE_II(MCDA):
    def __init__(self):
        self.df_dif_pref = 0        
        self.df_apf = 0        
        self.df_flows = 0        
        super().__init__()
    
    def decide(self, normalization_method=LinearMinMax_):
        """
        Method for decision making using PROMETHEE_II. 

        :param normalization_method: (default=LinearMinMax_) Must be defined when the problem data is different units,
        :type normalization_method: some value defined in NORMALIZATION_METHODS constant

        :returns: none value
        :rtype: none
        """         
        self._MCDA__set_normalization_method(normalization_method)
        self._MCDA__normalize()
        self.__dif_pref()
        self.__flows()
        self.__promethee_ii()

    def __flows(self):
        df_temp = pd.DataFrame({"alternatives":  list(self.df_original.iloc[:, 0]),
                                "Leaving Flow φ+":  list(self.df_apf.iloc[:, 1:].sum(axis=1)/(self.df_apf.shape[0]-1)),
                                "Flow φ-": list(self.df_apf.iloc[:, 1:].sum(axis=0)/(self.df_apf.shape[0]-1))
                                })
        
        df_temp["Net Flow φ"] = df_temp.iloc[:, 1] - df_temp.iloc[:, 2] 
        self.df_flows = df_temp

    
        
    def __dif_pref(self):
    
        # define shape of dif table
        n_rows = (self.df_normalized.shape[0]-1)*(self.df_normalized.shape[1]-1)
        df_temp = pd.DataFrame(0, index=np.arange(n_rows), columns=self.df_normalized.iloc[:, 1:].columns.values)
    
        labels = [None] * n_rows
    
        for c in range(1, self.df_normalized.shape[1]):
            r = 0
            for i in range(self.df_normalized.shape[0]):
                for j in range(self.df_normalized.shape[0]):            
                    if i != j: 
                        dif = self.df_normalized.iloc[i,c] - self.df_normalized.iloc[j,c]
                        if dif < 0:
                            dif = 0
                        df_temp.iloc[r,c-1] = dif * self.weights[c-1]
                        labels[r] = "(M" + str(i+1) + " x M" + str(j+1) + ")"
                        r = r + 1
                        if r >= n_rows:
                            r = 0

        # Aggregating Perform Function
        df_temp["APF"] = df_temp.iloc[:, 0:].sum(axis=1) / sum(self.weights) 
         
        alternatives = list(self.df_original.iloc[:, 0])

        df_apf = pd.DataFrame(0, index=np.arange(len(alternatives)), columns=alternatives)

        n_colsrows = self.df_original.shape[0]
        x = 0
        for i in range(0, n_colsrows):
            for j in range(0, n_colsrows):            
                if i != j:
                    df_apf.iloc[i, j] = df_temp.iloc[x, 4]
                    x = x + 1
                    
        self.df_dif_pref = pd.DataFrame(labels, columns=["CRITERIA"]).join(df_temp)
        self.df_weighted = self.df_dif_pref
        self.df_apf = pd.DataFrame(alternatives, columns=["APF"]).join(df_apf)


    def __promethee_ii(self):
        i = np.arange(1, len(self.df_original.index)+1, 1)
        df_ranking = self.df_flows.sort_values(by=["Net Flow φ"], ascending=False)
        df_ranking["rank"] = i
        
        self.df_decision = df_ranking.sort_index()


'''
#########################################################
# SCI-KIT-MCDA Library                                  #
# Author: Antonio Horta                                 #
# https://gitlab.com/cybercrafter/scikit-mcda           #
# Cybercrafter ® 2021                                   #
#########################################################
'''

import pandas as pd
import copy
from tabulate import tabulate

class DMUU:
    """
    Class: DMUU: Decision making under uncertainty
    """
    def __init__(self):
        self.df_original = 0
        self.df_calc = 0
        self.decision = []

    def dataframe(self, alt_data, alt_labels=None, state_labels=None):

        # define state labels if not exists
        if state_labels is None:
            state_labels = []
            for s in range(1, len(alt_data[0])+1):
                state_labels.append("S" + str(s))

        # define alternative labels if not exists
        if alt_labels is None:
            alt_labels = []
            for a in range(1, len(alt_data)+1):
                alt_labels.append("A" + str(a))

        df_data = pd.DataFrame(data=alt_data, columns=state_labels)
        df_data.insert(loc=0,
                    column='alternatives',
                    value=alt_labels)
        dfo = df_data

        self.df_original = copy.copy(dfo)
        self.df_calc = copy.copy(dfo)
        self.decision = []

    def minimax_regret(self):
        self.__calc("minimax-regret")
        self.decision_making("minimax-regret")
        return "minimax-regret"

    def maximax(self):
        self.__calc("maximax")
        self.decision_making("maximax")
        return "maximax"

    def maximin(self):
        self.__calc("maximin")
        self.decision_making("maximin")
        return "maximin"

    def laplace(self):
        self.__calc("laplace")
        self.decision_making("laplace")
        return "laplace"

    def hurwicz(self, hurwicz_coeficient=-1):
        self.__calc("hurwicz", hurwicz_coeficient)
        self.decision_making("hurwicz")
        return "hurwicz"

    def calc_clean(self):
        self.df_calc = copy.copy(self.df_original)
        return True

    '''
    FUNCTION:
    decision_making(dmuu_df, dmuu_criteria_list=[], hurwicz_coeficient=0.5):

    DESCRIPTION:
    Function that returns solutions DMUU from dmuu_dataframe or dmcu_dataframe_calculated of criteria listed in dmuu_criteria_list

    PARAMS:
    dmuu_df: dataframe with alteratives and states
    dmuu_criteria_list: list of criteria ["maximax", "maximin", "laplace", "minimax-regret", "hurwicz"]

    OUTPUT:
            list of dict         {"alternative": ,
                                "index": ,
                                "value": ,
                                "criteria":,
                                "result": [{"alternative": value}],
                                "type_dm": "DMUU",
                                'hurwicz_coeficient':
                                }
    '''

    def decision_making(self, dmuu_criteria_list=[]):

        # check dmuu_criteria_list
        dmuu_criteria_list = self.__check_dmuu_criteria_list(dmuu_criteria_list)

        result = []

        # make result
        for l in dmuu_criteria_list:
            cols = list(self.df_calc[l])
            for c in cols:
                if c[1] == 1:
                    hc = ""
                    if l in "hurwicz":
                        hc = c[2]
                    result.append({"alternative": self.df_calc.iloc[cols.index(c), 0],
                                "index": cols.index(c),
                                "value": c[0],
                                "criteria": l,
                                "result": dict(zip(self.df_calc["alternatives"], [i[0] for i in cols])),
                                "type_dm": "DMUU",
                                "hurwicz_coeficient": hc
                                })
        
        self.decision = result


    def __calc(self, dmuu_criteria_list=[], hurwicz_coeficient=-1):

        df = self.df_calc

        if "minimax-regret" in dmuu_criteria_list:
            list_minimax_regret_max = self.df_original.iloc[:, 1:].max(axis=0)
            df_minimax_regret_temp = self.df_original.iloc[:, 1:] - list_minimax_regret_max
            # now apply maximin
            df_maximin_r = df_minimax_regret_temp.min(axis=1)
            result_maximin_r = df_maximin_r.max(axis=0)
            column_maximin_r = []
            for x in df_maximin_r.values:
                if x == result_maximin_r:
                    column_maximin_r.append((x * -1, 1))
                else:
                    column_maximin_r.append((x * -1, 0))
            df['minimax-regret'] = column_maximin_r

        if "maximax" in dmuu_criteria_list:
            df_maximax = self.df_original.iloc[:, 1:].max(axis=1)
            result_maximax = df_maximax.max(axis=0)
            column_maximax = []
            for x in df_maximax.values:
                if x == result_maximax:
                    column_maximax.append((x, 1))
                else:
                    column_maximax.append((x, 0))
            df['maximax'] = column_maximax

        if "maximin" in dmuu_criteria_list:
            df_maximin = self.df_original.iloc[:, 1:].min(axis=1)
            result_maximin = df_maximin.max(axis=0)
            column_maximin = []
            for x in df_maximin.values:
                if x == result_maximin:
                    column_maximin.append((x, 1))
                else:
                    column_maximin.append((x, 0))
            df['maximin'] = column_maximin

        if "laplace" in dmuu_criteria_list:
            df_laplace = self.df_original.iloc[:, 1:].mean(axis=1)
            result_laplace = df_laplace.max(axis=0)
            column_laplace = []
            for x in df_laplace.values:
                if x == result_laplace:
                    column_laplace.append((x, 1))
                else:
                    column_laplace.append((x, 0))
            df['laplace'] = column_laplace

        if "hurwicz" in dmuu_criteria_list:

            # check hurwicz_coeficient is ok. if not default is 0.5
            if self.__check_hurwicz_coeficient(hurwicz_coeficient) is False:
                hurwicz_coeficient = 0.5

            df_hurwicz = self.df_original.iloc[:, 1:].max(axis=1) * hurwicz_coeficient + self.df_original.iloc[:, 1:].min(axis=1) * (1 - hurwicz_coeficient)
            result_hurwicz = df_hurwicz.max(axis=0)
            column_hurwicz = []
            for x in df_hurwicz.values:
                if x == result_hurwicz:
                    column_hurwicz.append((x, 1, hurwicz_coeficient))
                else:
                    column_hurwicz.append((x, 0, hurwicz_coeficient))
            df['hurwicz'] = column_hurwicz

        self.df_calc = df


    '''
    FUNCTION:
    check_hurwicz_coeficient(hurwicz_coeficient):

    DESCRIPTION:
    Check if hurwicz_coeficient is >= 0 and <= 1
    If not, put default 0.5

    OUTPUT:
        True or False
    '''


    def __check_hurwicz_coeficient(self, hurwicz_coeficient):
        result = True
        if type(hurwicz_coeficient) != float:
            if hurwicz_coeficient < 0 or hurwicz_coeficient > 1:
                result = False
        return result


    '''
    FUNCTION:
    check_dmuu_criteria_list(dmuu_criteria_list=[]):

    DESCRIPTION:
    Check and transform if dmuu_criteria_list is a list of correct dmuu criteria
    and defines default criteria for DMUU

    OUTPUT:
        dmuu_criteria_list transformed or default
    '''


    def __check_dmuu_criteria_list(self, dmuu_criteria_list=[]):

        # defines default criteria for DMUU
        default_list = ["maximax", "maximin", "laplace", "minimax-regret", "hurwicz"]

        if type(dmuu_criteria_list) != list:
            dmuu_criteria_list = [dmuu_criteria_list]

        for c in dmuu_criteria_list:
            if c not in default_list:
                dmuu_criteria_list.remove(c)

        if dmuu_criteria_list == []:
            dmuu_criteria_list = default_list

        return dmuu_criteria_list

    def pretty_decision(self, tablefmt='psql'):
        return tabulate(self.decision, headers='keys', tablefmt=tablefmt)

    def pretty_calc(self, tablefmt='psql'):
        return tabulate(self.df_calc, headers='keys', tablefmt=tablefmt)

    def pretty_original(self, tablefmt='psql'):
        return tabulate(self.df_original, headers='keys', tablefmt=tablefmt)



from .mcda import MCDA, pd, np

class WASPAS(MCDA):
    def __init__(self):
        super().__init__()        
    
    def decide(self, lambda_=0.5, normalization_method=None):
        """
        Method for decision making using WASPAS. 

        :param lambda_: (default=0.5) Must be defined for tunning purpose,
        :type lambda_: float

        :param normalization_method: (default=None) Must be defined when the problem data is different units,
        :type normalization_method: some value defined in NORMALIZATION_METHODS constant

        :returns: none value
        :rtype: none
        """   
        if self.__check_lambda(lambda_) is True:
            self._MCDA__set_normalization_method(normalization_method)
            self._MCDA__normalize()
            self._MCDA__weighting_from_normalized()
            self.__waspas(lambda_)

    def __waspas(self, lambda_):
        w_df = pd.DataFrame(self.df_weighted.iloc[:, 1:])       
       
        q_wsm = pd.DataFrame(w_df.sum(axis=1), columns=["WSM"]) * lambda_
        q_wpm = pd.DataFrame(w_df.prod(axis=1), columns=["WPM"]) * (1 - lambda_) 

        waspas = q_wsm.iloc[:,0] + q_wpm.iloc[:,0]
        label = "WASPAS (λ " + str(lambda_) + " )" 
        waspas_df = pd.DataFrame(waspas, columns=[label]) 

        i = np.arange(1, len(self.df_original.index)+1, 1)
        df_ranking = pd.DataFrame(self.df_original.iloc[:,0]).join(waspas_df).sort_values(by=[label], ascending=False)
        df_ranking["rank"] = i

        self.df_decision = df_ranking.sort_index()

    def __check_lambda(self, lambda_):
        result = False
        if type(lambda_) == float:
            if (lambda_ >= 0 and lambda_ <= 1):
                result = True
            else:
                raise ValueError("Lambda must be a value between 0 and 1!")                
        return result


from .mcda import MCDA, pd, np, tabulate

class ELECTRE_I(MCDA):
    def __init__(self):
        super().__init__()     
        self.df_concordance_matrix: 0
        self.df_discordance_matrix: 0
        self.df_credibility_matrix: 0
   
    def decide(self, c=1, d=0, normalization_method=None):
        """
        Method for decision making using ELECTRE I. 

        :param c: (optional) concordance index. Default 1 is the strict case  
        :type c: c > 0.5 up to 1

        :param d: (optional) discordance index. Deafault 0 is the strict case  
        :type c: c < 0.5 up to 0

        :param normalization_method: (default=None) Must be defined when the problem data is different units,
        :type normalization_method: some value defined in NORMALIZATION_METHODS constant

        :returns: print
        :rtype: string
        """          
        if self.__check_cd(c, d) is True:
            self._MCDA__set_normalization_method(normalization_method)
            self._MCDA__normalize()
            self.__make_concordance_matrix()
            self.__make_discordance_matrix()
            self.__electre_i(c, d)
            return "electre_i"

    def __make_concordance_matrix(self):
        # define shape of c_matrix
        n_rows = (self.df_normalized.shape[0])
        alternatives = list(self.df_normalized.iloc[:, 0])
        df_concordance_mtx = pd.DataFrame(0, index=np.arange(n_rows), columns=alternatives)
        df_concordance_mtx.insert(loc=0, column='Concordance MTX', value=alternatives)

        W = sum(self.weights)

        for alt in range(n_rows):
            for i in range(n_rows):
                if i != alt:
                    Cab = 0
                    for col in range(1, self.df_normalized.shape[1]):
                        a = self.df_normalized.iloc[alt, col]
                        b = self.df_normalized.iloc[i, col]
                        if a >= b:
                            Cab = Cab + self.weights[col-1]
                        else:
                            Cab = Cab + 0
                    Cab = Cab/W
                    df_concordance_mtx.iloc[alt, i+1] = Cab

        self.df_concordance_matrix = df_concordance_mtx
    
    def __make_discordance_matrix(self):
        # define shape of c_matrix
        n_rows = (self.df_normalized.shape[0])
        alternatives = list(self.df_normalized.iloc[:, 0])
        df_discordance_mtx = pd.DataFrame(0, index=np.arange(n_rows), columns=alternatives)
        df_discordance_mtx.insert(loc=0, column='Discordance MTX', value=alternatives)

        S = list(self.df_normalized.iloc[:, 1:].max(axis=0) - self.df_normalized.iloc[:, 1:].min(axis=0))
      
        for alt in range(n_rows):
            for i in range(n_rows):
                if i != alt:
                    Dab = []
                    count_negatives = 0
                    for col in range(1, self.df_normalized.shape[1]):
                        a = self.df_normalized.iloc[alt, col]
                        b = self.df_normalized.iloc[i, col]
                        d = (b - a)/S[col-1]
                        Dab.append(d)
                        if d < 0:
                            count_negatives = count_negatives + 1
                    if count_negatives == n_rows:
                        Dab = 0
                    else:
                        Dab = max(Dab)
                    df_discordance_mtx.iloc[alt, i+1] = Dab
        
        self.df_discordance_matrix = df_discordance_mtx

    def __electre_i(self, c, d):

        # init credibility matrix
        self.df_credibility_matrix = self.df_concordance_matrix.rename(columns={'Concordance MTX': 'OUTRANKING'})

        n_rows = self.df_normalized.shape[0]
        dominance_relationship = []
        for i in range(n_rows):
            prevalent = 0
            for j in range(n_rows):
                c_cell = self.df_concordance_matrix.iloc[i, j+1]
                d_cell = self.df_discordance_matrix.iloc[i, j+1]
                if i != j:
                    if c_cell >= c and d_cell <= d:
                        self.df_credibility_matrix.iloc[i, j+1] = str(1)
                        prevalent = prevalent + 1     
                    else:
                        self.df_credibility_matrix.iloc[i, j+1] = str(0)
                else:
                    self.df_credibility_matrix.iloc[i, j+1] = "-"
            if prevalent > 0:
                dominance_relationship.append("prevalent")
            else:
                dominance_relationship.append("subdued")

        df_dominance_relationship = pd.DataFrame({"Alternatives": self.df_concordance_matrix.iloc[:, 1:].columns,
                                                  "Outranking": dominance_relationship})

        self.df_decision = df_dominance_relationship

    def __check_cd(self, c, d):
        if c > 0.5 and d < 0.5:
            return True
        else:
            raise ValueError("Wrong concordance (c > 0.5) or discrdance (d < 0.5) indexes")

    def get_graph(self):
        """
        Function returns nodes and edges to draw electre I graph.

        :returns: {"nodes": list of nodes_id, 
                   "edges": list of edges in generic format {"source": node_id ,"target": node_id}, 
                   "edges_networkx": list of edges in networkx format
                   } 
        :rtype: dict
        """
        n_rows = self.df_credibility_matrix.shape[0]
        edges = []
        edges_networkx = []
        nodes = list(self.df_credibility_matrix.iloc[:, 0])
        for a in range(n_rows):
            for b in range(n_rows):
                if self.df_credibility_matrix.iloc[a, b+1] == "1":
                    edges.append({"source": self.df_credibility_matrix.iloc[a, 0],"target": self.df_credibility_matrix.columns[b+1]})
                    edges_networkx.append((self.df_credibility_matrix.iloc[a, 0],self.df_credibility_matrix.columns[b+1]))
        
        return {"nodes": nodes, "edges": edges, "edges_networkx": edges_networkx}

    def pretty_concordance_matrix(self, tablefmt='psql'):
        """
        Print a pretty table about specific attribute.

        :param tablefmt: (default=psql) format of output 
        :type: string constant: 'html', 'latex' or 'psql'

        :returns: print
        :rtype: string

        """        
        return tabulate(self.df_concordance_matrix, headers='keys', tablefmt=tablefmt)
            
    def pretty_discordance_matrix(self, tablefmt='psql'):
        """
        Print a pretty table about specific attribute.

        :param tablefmt: (default=psql) format of output 
        :type: string constant: 'html', 'latex' or 'psql'

        :returns: print
        :rtype: string

        """        
        return tabulate(self.df_discordance_matrix, headers='keys', tablefmt=tablefmt)

    def pretty_credibility_matrix(self, tablefmt='psql'):
        """
        Print a pretty table about specific attribute.

        :param tablefmt: (default=psql) format of output 
        :type: string constant: 'html', 'latex' or 'psql'

        :returns: print
        :rtype: string

        """        
        return tabulate(self.df_credibility_matrix, headers='keys', tablefmt=tablefmt)



from .mcda import MCDA, pd, np
from .topsis import TOPSIS
from .wpm import WPM
from .wsm import WSM
from .vikor import VIKOR
from .promethee_ii import PROMETHEE_II
from .electre_ii import ELECTRE_II
from .waspas import WASPAS


class ENSEMBLE(MCDA):
    """
    The concept of ensemble class came from machine learning methods to combine the predictions 
    of several base estimators built with a given learning algorithm in order to improve generalizability
    or robustness over a single estimator.
    """
    def __init__(self):
        super().__init__()        
        self.df_decision: 0
    
    def ranking_by_voting(self, list_mcda_intances, voting='soft'):  
        """
        The idea behind the ranking_by_voting Method is to combine conceptually different MCDA Raking Methods and 
        use a majority vote (hard vote) or the average of preferences (soft vote) to estabilish a ranking. 
        
        It can be useful for a set of equally well performing model in order to balance out their individual weaknesses.

        In majority voting, the priority/rank achieved for a particular alternative is the priority/rank 
        that represents the majority (mode) achieved by each individual MCDA method.

        E.g., if the rank for a given problem is

                            PROMETHEE II  TOPSIS   WSM   WPM    Majority
        Alternative 1 ->              1º      2º     2º   2º          2º
        Alternative 2 ->              3º      3º     1º   3º          3º
        Alternative 3 ->              2º      1º     3º   1º          1º
        
        the ranking_by_voting (with 'hard') would rank the problem as “A1: 2º, A2: 3º and A3: 1º” based on the majority.

        In the cases of a tie, the ranking_by_voting will set "a tie" as result.      

        The hard method allows duplicated priority
      
        :param list_mcda_intances: (default=None) Must be defined when the problem data is different units,
        :type list_mcda_intances: some value defined in NORMALIZATION_METHODS constant

        :param voting: (default='soft') type of voting 'hard' or 'soft'
        :type voting: string 'hard' or 'soft'

        :returns: none value
        :rtype: none
        """   
        if self.__check_mcda_method(list_mcda_intances) is True:
            if self.__check_voting_method(voting) == 'soft':
                self.__soft_voting(list_mcda_intances)
                self.__verdict()
            else:
                self.__hard_voting(list_mcda_intances)
        
            return self.df_decision

    def __check_voting_method(self, voting):
        if voting in ["hard", "soft"]:
            return voting
        else:
            return "soft"

    def __check_mcda_method(self, list_mcda_intances):
        for mcdm in list_mcda_intances:
            if type(mcdm) not in [TOPSIS, PROMETHEE_II, WASPAS, WSM, WPM, VIKOR, ELECTRE_II]:
              raise ValueError("list_mcda_intances must be a list of instances of MCDA RANKING methods (TOPSIS, PROMETHEE_II, WASPAS, WSM, WPM, VIKOR, ELECTRE_II)")        
        return True

    def __verdict(self):
        i = np.arange(1, len(self.df_decision.index)+1, 1)
        df_ranking = self.df_decision.sort_values(by=["voting"], ascending=False)
        df_ranking["rank"] = i
        self.df_decision = df_ranking.sort_index()

    def __soft_voting(self, list_mcda_intances):
        df_soft = pd.DataFrame(np.zeros((list_mcda_intances[0].df_decision.shape[0], len(list_mcda_intances))))
        column = 0
        for m in list_mcda_intances:
            if type(m) == VIKOR: 
                df_soft[column] = 1 - m.df_decision.iloc[:, -2] # VIKOR need to be inverted for sort correctly            
            else:
                df_soft[column] = m.df_decision.iloc[:, -2]
            df_soft = df_soft.rename(columns={column: type(m).__name__})
            column = column + 1

        df_soft["voting"] = df_soft.iloc[:,:].mean(axis=1)
        
        df_soft = pd.DataFrame(list_mcda_intances[0].df_decision.iloc[:, 0]).join(df_soft)
        
        self.df_decision = df_soft

    def __hard_voting(self, list_mcda_intances):
        df_hard = pd.DataFrame(np.zeros((list_mcda_intances[0].df_decision.shape[0], len(list_mcda_intances))))
        column = 0
        for m in list_mcda_intances:
            if type(m) == VIKOR: # filter * in VIKOR compromise conditions
                vikor_list = str(m.df_decision.iloc[:, -1].values).replace("*", "").replace("[", "").replace("]", "").replace("'", "").replace(" ", "")
                vikor_list = np.array(list(vikor_list), dtype=int)
                df_hard[column] = vikor_list
            else:
                df_hard[column] = m.df_decision.iloc[:, -1]
            df_hard = df_hard.rename(columns={column: type(m).__name__})
            column = column + 1

        # Get a tie
        a_tie = df_hard.iloc[:,:].agg(lambda x: "a tie" if x.value_counts().values[0] in x.value_counts().values[1:] else x.value_counts().values[0] , axis=1).values.tolist()

        df_hard["std"] = df_hard.iloc[:,:].std(axis=1)
        df_hard["voting"] = df_hard.iloc[:,:].agg(lambda x: x.value_counts().index[0], axis=1)
        df_hard["rank"] = df_hard.iloc[:,:].agg(lambda x: x.value_counts().index[0], axis=1)

        # Set a tie
        for t in a_tie:
            if t == "a tie":
                df_hard.iloc[a_tie.index(t), -2] = "a tie"
                df_hard.iloc[a_tie.index(t), -1] = "a tie"

        df_hard = pd.DataFrame(list_mcda_intances[0].df_decision.iloc[:, 0]).join(df_hard)
        
        self.df_decision = df_hard



from .mcda import MCDA, pd, np, Vector_, tabulate

class TOPSIS(MCDA):
    def __init__(self):
        self.df_pis = []
        self.df_nis = []
        self.df_closeness = 0
        super().__init__()        
    
    def decide(self, normalization_method=Vector_):        
        """
        Method for decision making using TOPSIS. 

        :param normalization_method: (default=Vector_) Must be defined when the problem data is different units,
        :type normalization_method: some value defined in NORMALIZATION_METHODS constant

        :returns: none value
        :rtype: none
        """   
        self._MCDA__set_normalization_method(normalization_method)      
        self._MCDA__normalize(True)
        self._MCDA__weighting_from_normalized()
        self.__xis()
        self.__topsis()
         
    def __xis(self):
        pis = pd.DataFrame(self.df_weighted.iloc[:, 1:] * self.signals).max(axis=0) * self.signals
        self.df_pis = pis
        nis = pd.DataFrame(self.df_weighted.iloc[:, 1:] * self.signals).min(axis=0) * self.signals
        self.df_nis = nis

    def __topsis(self):
        dp = np.sqrt(self.df_weighted.iloc[:, 1:].sub(self.df_pis).pow(2).sum(axis=1))
        dn = np.sqrt(self.df_weighted.iloc[:, 1:].sub(self.df_nis).pow(2).sum(axis=1))
        closeness = pd.DataFrame(dn.div(dp+dn), columns=["performance score"])
        i = np.arange(1, len(self.df_original.index)+1, 1)
        df_ranking = pd.DataFrame(self.df_original.iloc[:,0]).join(closeness).sort_values(by=["performance score"], ascending=False)
        df_ranking["rank"] = i

        self.df_decision = df_ranking.sort_index()

    def pretty_Xis(self, tablefmt='psql'):
        """
        Print a pretty table about positive and negative ideal solutions attributes (df_pis and df_nis).

        :param tablefmt: (default=psql) format of output 
        :type: string constant: 'html', 'latex' or 'psql'

        :returns: print
        :rtype: string

        """        
        return tabulate(pd.DataFrame(self.df_pis, columns=['PIS']).join(pd.DataFrame(self.df_nis, columns=["NIS"])).T, headers='keys', tablefmt=tablefmt)



from .electre_i import ELECTRE_I, np, pd

class ELECTRE_II(ELECTRE_I):
    def __init__(self):
        super().__init__()     
        self.pure_concordance: 0
        self.pure_discordance: 0
   
    def decide(self, c=1, d=0, normalization_method=None):
        """
        Method for decision making using ELECTRE II. It results in a ranking of alternatives 
        by median of pure concordance x pure discordance rankings

        :param c: (optional) concordance index. Default 1 is the strict case  
        :type c: c > 0.5 up to 1

        :param d: (optional) discordance index. Deafault 0 is the strict case  
        :type c: c < 0.5 up to 0

        :param normalization_method: (default=None) Must be defined when the problem data is different units,
        :type normalization_method: some value defined in NORMALIZATION_METHODS constant

        :returns: print
        :rtype: string
        """          
        if self._ELECTRE_I__check_cd(c, d) is True:
            self._MCDA__set_normalization_method(normalization_method)
            self._MCDA__normalize()
            self._ELECTRE_I__make_concordance_matrix()
            self._ELECTRE_I__make_discordance_matrix()
            self._ELECTRE_I__electre_i(c, d)
            self.__electre_ii()
            return "electre_ii"

    def __electre_ii(self):
        self.__calc_pure_concordance()
        self.__calc_pure_discordance()
        self.__median()
        
    def __calc_pure_concordance(self):
    
        sum_row = np.array(self.df_concordance_matrix.iloc[:, 1:].sum(axis=1).to_list())
        sum_col = np.array(self.df_concordance_matrix.iloc[:, 1:].sum(axis=0).to_list())
        self.pure_concordance = pd.DataFrame(sum_row - sum_col, columns=["pure concordance"])
        
        i = np.arange(1, len(self.df_original.index)+1, 1)
        df_ranking = pd.DataFrame(self.df_original.iloc[:,0]).join(self.pure_concordance).sort_values(by=["pure concordance"], ascending=False)
        df_ranking["rank concordance"] = i

        self.pure_concordance = df_ranking.sort_index()
        
    def __calc_pure_discordance(self):
        
        sum_row = np.array(self.df_discordance_matrix.iloc[:, 1:].sum(axis=1).to_list())
        sum_col = np.array(self.df_discordance_matrix.iloc[:, 1:].sum(axis=0).to_list())
        self.pure_discordance = pd.DataFrame(sum_row - sum_col, columns=["pure discordance"])
    
        i = np.arange(1, len(self.df_original.index)+1, 1)
        df_ranking = pd.DataFrame(self.df_original.iloc[:,0]).join(self.pure_discordance).sort_values(by=["pure discordance"], ascending=True)
        df_ranking["rank discordance"] = i

        self.pure_discordance = df_ranking.sort_index()

    def __median(self):
        ranks = self.pure_concordance.join(self.pure_discordance.iloc[:,-1])
        median = pd.DataFrame(ranks.iloc[:,-2:].median(axis=1), columns=["median scaled"]) / ranks.iloc[:,-2:].median(axis=1).sum(axis=0)
        final_rank = pd.DataFrame(ranks.iloc[:,-2:].median(axis=1), columns=["rank"])
        self.df_decision = self.pure_concordance.join(self.pure_discordance.iloc[:,1:]).join(median).join(final_rank)



from .mcda import MCDA, pd, np, MIN, MAX

class VIKOR(MCDA):
    """
    *Višekriterijumska Optimizacija I Kompromisno Rješenje (VIKOR)*

    References:

    - OPRICOVIC, S. Fuzzy VIKOR with an application to water resources planning. Expert Systems with Apliccations, v. 38, n. 10, p. 12983- 12990, 2011.
    - Opricovic, S. Tzeng, G.-H. (2007). Extended VIKOR method in comparison with outranking methods. European Journal of Operational Research, 178(2): 514-529.
    - OPRICOVIC, S. TZENG, G.. Compromisse solution by MCDM methods: a comparative analysis of VIKOR and TOPSIS. European Journal of Operational Research, v. 16, p. 445-455, 2004.
    - OPRICOVIC, S. TZENG, G.. Multicriteria planning of post-earthquake sustainable reconstruction. Computer-Aided Civil and Infrastructure Engineering, v. 17, p. 211-220, 2002.
    """
    def __init__(self):
        super().__init__()
        self.x_best = []
        self.x_worst = []
        self.df_Si_Ri = []        
        self.S_minmax = (0, 0)
        self.R_minmax = (0, 0)
        self.df_Qi = []        
        self.Conditions = {"C1": True, "C2": True}
        
    def decide(self, v=0.5):  
        """
        Method for decision making using VIKOR. 
        Multi-criteria Optimization AND compromise solution

        :param v: (default=0.5) Where v the weight for the strategy of maximum group utility and 
                  1 - v is the weight of the individual regret. 
                  Usually v = 0.5 and when v > 0.5, the index of Qj will tend to majority agreement 
                  and clearly when v < 0.5, the index of Qj will indicate majority negative attitude.

        :returns: Conditions C1 - Acceptable Advantage and C2 - Accetable Stability in Decision Making
        :rtype: dict
        """   
        self.__calc_X()
        self.__calc_Si_Ri()
        self.__calc_Qi(v)
        self.__vikor()

        return self.Conditions

    def __calc_X(self):
        max_ = list(self.df_original.iloc[:, 1:].max(axis=0)) 
        min_ = list(self.df_original.iloc[:, 1:].min(axis=0))
       
        self.x_best = []
        self.x_worst = []        
    
        i = 0
        for s in self.signals:
            if s > 0:
                self.x_best.append(max_[i])
                self.x_worst.append(min_[i])
            else:
                self.x_best.append(min_[i])
                self.x_worst.append(max_[i])
            i = i + 1

    def __calc_Si_Ri(self):

        self.df_Si_Ri = self.df_original
        
        n_rows, n_cols = self.df_original.shape
        
        Si = 0
        for i in range(n_rows): 
            for j in range(n_cols-1): 
                if (self.x_best[j] - self.x_worst[j]) == 0: # prevent div by zero
                    Si = 0
                else:
                    Si = self.weights[j] * (  (self.x_best[j] - float(self.df_original.iloc[i, j+1]))  / (self.x_best[j] - self.x_worst[j]) )    
                self.df_Si_Ri.iloc[i, j+1] = Si
        
        self.df_Si_Ri["Si"] = self.df_Si_Ri.sum(axis=1) 
        self.df_Si_Ri["Ri"] = self.df_Si_Ri.iloc[:, 1:-1].max(axis=1) 

        self.S_minmax = (self.df_Si_Ri.iloc[:, -2].min(axis=0), self.df_Si_Ri.iloc[:, -2].max(axis=0))
        self.R_minmax = (self.df_Si_Ri.iloc[:, -1].min(axis=0), self.df_Si_Ri.iloc[:, -1].max(axis=0))

    def __calc_Qi(self, v):

        self.df_Qi = self.df_Si_Ri
        
        n_rows, n_cols = self.df_original.shape
        
        Qi = []
        for i in range(n_rows): 
            Qi.append(v * ((self.df_Si_Ri.iloc[i, -2] - self.S_minmax[0])/(self.S_minmax[1] - self.S_minmax[0])) +  ((1-v) * ((self.df_Si_Ri.iloc[i, -1] - self.R_minmax[0])/(self.R_minmax[1] - self.R_minmax[0]))))
        
        self.df_Qi["Qi"] = Qi
        self.df_normalized = self.df_Qi
        self.df_weighted = self.df_Qi

    def __vikor(self):
        i = np.arange(1, len(self.df_original.index)+1, 1)
        df_ranking = pd.DataFrame(self.df_original.iloc[:,0]).join(self.df_Qi.iloc[:,-3:]).sort_values(by=["Qi"], ascending=True)
        df_ranking["rank"] = i
        
        self.df_decision = df_ranking.sort_index()
        self.__check_compromise()

    def __check_compromise(self):
        # Check Conditions C1 and C2
        QA1 = 0
        RA1 = 0
        SA1 = 0
        QA2 = 0
        DQ = 1 / (self.df_original.shape[0] - 1)
        for r in range(self.df_decision.shape[0]):
            if self.df_decision.iloc[r, 4] == 1:
                QA1 =  self.df_decision.iloc[r, 3]
                RA1 =  self.df_decision.iloc[r, 2]
                SA1 =  self.df_decision.iloc[r, 1]
            if self.df_decision.iloc[r, 4] == 2:
                QA2 =  self.df_decision.iloc[r, 3]
                RA2 =  self.df_decision.iloc[r, 2]
                SA2 =  self.df_decision.iloc[r, 1]

        # Check C1 - Acceptable Advantage
        if QA2 - QA1 >= DQ:
            self.Conditions["C1"] = True
        else:
            self.Conditions["C1"] = False
            # Just Q(A^(M)) - Q(A1) < DQ. where max M
            for r in range(self.df_decision.shape[0]):
                if self.df_decision.iloc[r, 3] - QA1 < DQ:
                    self.df_decision.iloc[r, 4] = str(self.df_decision.iloc[r, 4]) + "*"

        # check C2 - Accetable Stability in Decision Making
        # Alterative A1 must also be the best ranked by S or/and R
        S_min = self.df_decision.iloc[:, 1].min()
        R_min = self.df_decision.iloc[:, 2].min()
        if SA1 == S_min or RA1 == R_min:
            self.Conditions["C2"] = True
        else:
            self.Conditions["C2"] = False
            # Just A1 and A2
            for r in range(self.df_decision.shape[0]):
                if self.df_decision.iloc[r, 4] == 1 or self.df_decision.iloc[r, 4] == 2:
                    self.df_decision.iloc[r, 4] = str(self.df_decision.iloc[r, 4]) + "*"


MAX = +1
MIN = -1

LinearMinMax_ = "LinearMinMax"
LinearMax_ = "LinearMax"
LinearSum_ = "LinearSum"
Vector_ = "Vector"
EnhancedAccuracy_ = "EnhancedAccuracy"
Logarithmic_ = "Logarithmic"

NORMALIZATION_METHODS = [LinearMinMax_,
                         LinearMax_,
                         LinearSum_,
                         Vector_,
                         EnhancedAccuracy_,
                         Logarithmic_]
                         
SAATY_SCALE = [1, 2 ,3 ,4 ,5 ,6 ,7 ,8 ,9, 1/2, 1/3, 1/4, 1/5, 1/6, 1/7, 1/8, 1/9]

SAATY_RI = [0, 0, 0.58, 0.9, 1.12, 1.24, 1.32, 1.41, 1.45, 1.49, 1.51, 1.53, 1.56, 1.57, 1.59]


'''
#########################################################
# SCI-KIT-MCDA Library                                  #
# Author: Antonio Horta                                 #
# https://gitlab.com/cybercrafter/scikit-mcda           #
# Cybercrafter ® 2021                                   #
#########################################################
'''

import pandas as pd
import numpy as np
import copy
from tabulate import tabulate
from .constants import *
import math

class MCDA:
    """
    Class: MCDA: Multi-Criteria Decision Aid
    """
    def __init__(self):
        self.df_original = 0
        self.weights = []
        self.signals = []
        self.normalization_method = None
        self.df_normalized = 0
        self.df_weighted = 0
        self.df_decision = []

    def dataframe(self, alt_data, alt_labels=None, criteria_labels=None):
        """
        Set data for MCDM problem and save it into attribute: df_original.
        
        :param alt_data: They are the values of each criteria (columns) for each alternative (rows),
        :type alt_data: list of list
        
        :param alt_labels: (Optional) Requires a list of labels for alternatives. If not provided, labels are automatically defined,
        :type alt_labels: list
        
        :param criteria_labels: (Optional) Requires a list of labels for criteria. If not provided, labels are automatically defined. Its very important that order of criteria are from less to more important due to some algorithms requirements for weights definition.
        :type criteria_labels: list
        
        :returns: no value
        :rtype: none
        """

        # define criteria labels if not exists
        if criteria_labels is None:
            criteria_labels = []
            for s in range(1, len(alt_data[0])+1):
                criteria_labels.append("C" + str(s))

        # define alternative labels if not exists
        if alt_labels is None:
            alt_labels = []
            for a in range(1, len(alt_data)+1):
                alt_labels.append("A" + str(a))

        

        df_data = pd.DataFrame(data=alt_data, columns=criteria_labels)
        df_data.insert(loc=0,
                    column='alternatives',
                    value=alt_labels)
        dfo = df_data

        self.df_original = copy.copy(dfo)
        self.df_calc = copy.copy(dfo)
        self.decision = []

    def set_weights_manually(self, weights, free=False):
        """
        Set data weights manually for MCDM problems and save it into attribute: weights.
        
        :param weights: The numeric values of weights for each criteria (columns),
        :type weights: list
        
        :param free: (default=False) If False, checks if weight value is a float between 0 and 1 and the sum is 100%,
        :type free: boolean
        
        :returns: no value
        :rtype: none
        """
        if self.__check_weights(weights, free) is True:
            self.weights = weights
        else:
            raise ValueError("Invalid weights! Each weight must be float between 0 and 1 and same number of criteria")

    def set_weights_by_ranking_A(self):
        """
        Set weights for MCDM problems and save it into attribute: weights.
        The criteria must be ordered by importance c1 > c2 > c3 ... 
        Wj = {1 / rj} / {∑[1 / rk], where k range is 1 to n}

        :returns: no value
        :rtype: none
        """
        n = len(self.df_original.columns) - 1
        r = np.arange(1, n+1, 1).tolist()
        W = []
        b = 0
        for k in r: # b = ∑[1 / rk]
            b = b + (1 / k)
        for j in r: # Wj = {1 / rj} / b
            W.append((1 / j) / b)
        
        self.weights = W

    def set_weights_by_ranking_B(self):
        """
        Set weights for MCDM problems and save it into attribute: weights.
        The criteria must be ordered by importance c1 > c2 > c3 ... 
        Wj = {n -rj + 1} / {∑[ n - rk + 1], where k range is 1 to n}

        :returns: no value
        :rtype: none
        """
        n = len(self.df_original.columns) - 1
        r = np.arange(1, n+1, 1).tolist()
        W = []
        b = 0
        for k in r: # b = ∑[ n - rk + 1]
            b = b + (n - k + 1)
        for j in r: # Wj = {n -rj + 1} / b
            W.append((n - j + 1) / b)
        
        self.weights = W

    def set_weights_by_ranking_C(self):
        """
        Set weights for MCDM problems and save it into attribute: weights.
        The criteria must be ordered by importance c1 > c2 > c3 ... 
        Wj = {1 / n} * {∑[ 1 / K ], where K range is j to n}

        :returns: no value
        :rtype: none
        """
        n = len(self.df_original.columns) - 1
        r = np.arange(1, n+1, 1)
        r = r[::-1].tolist()

        W = []
        b = 0
        a = 1 / n
        for k in r: # b = ∑[ 1 / K ]
            b = b + (1 / k)
            W.insert(0, a * b)
        
        self.weights = W

    def set_weights_by_ranking_B_POW(self, P=0):
        """
        Set weights for MCDM problems and save it into attribute: weights.
        The criteria must be ordered by importance c1 > c2 > c3 ... 
        Wj = {n -rj + 1}^P / {∑[ n - rk + 1]^P, where k range is 1 to n}
        
        :param P: (default=0) 0 results in the same weight for all criteria,
        :type P: int

        :returns: no value
        :rtype: none
        """
        n = len(self.df_original.columns) - 1
        r = np.arange(1, n+1, 1).tolist()
        W = []
        b = 0
        for k in r: # b = ∑[ n - rk + 1]
            b = b + pow((n - k + 1),P)
        for j in r: # Wj = {n -rj + 1} / b
            W.append(pow((n - j + 1), P) / b)
  
        self.weights = W

    def set_weights_by_entropy(self, normalization_method_for_entropy=LinearSum_):
        """
        Set weights by entropy for MCDM problems and save it into attribute: weights.
        IMPORTANT: signals must have to be set before weights

        :param normalization_method_for_entropy: (default=LinearSum_) 0 results in the same weight for all criteria,
        :type normalization_method_for_entropy: some value defined in NORMALIZATION_METHODS constant

        :returns: no value
        :rtype: none
        """

        # check if signal are defined before weigths by entropy.   They are required for normalization
        if self.signals == []:
            raise ReferenceError("Signals must have to be set before weights to use entropy")

        # Save current norm method
        current_norm_method = self.normalization_method

        # appy new norm method for entropy and apply
        self.__set_normalization_method(normalization_method_for_entropy)

        self.__normalize()
        x, y = self.df_normalized.iloc[:,1:].shape

        h = 1 / math.log(x)

        entropies = -h * (self.df_normalized.iloc[:, 1:] * np.log(self.df_normalized.iloc[:, 1:])).sum(axis=0)

        distances = pd.DataFrame(1 - entropies)
        distances_sum = distances.iloc[:,:].sum(axis=0)

        # Weights vector
        W = distances/distances_sum.values

        w = []
        for i in W.values.tolist():
            w.append(i[0])
        
        # set entropy weights
        self.weights = w

        #back current nor method
        self.normalization_method = current_norm_method

    def set_weights_by_AHP(self, saaty_preference_matrix):
        """
        Set weights by AHP for MCDM problems and save it into attribute: weights.

        e.g.
                                              # C1   C2     C3   C4 
        w_AHP = electre_i.set_weights_by_AHP([[  1,    4,    5,   7],   # C1
                                              [1/4,    1,    3,   5],   # C2
                                              [1/5,  1/3,    1,   3],   # C3
                                              [1/7,  1/5,  1/3,   1]])  # C4
    
        :param saaty_preference_matrix: matrix of preferences of values in Saaty scale.
        :type saaty_preference_matrix: list

        :returns: no value
        :rtype: none
        """
        if self.__check_AHP_matrix(saaty_preference_matrix) is True:  
            saaty_preference_matrix = pd.DataFrame(saaty_preference_matrix)
            priority_vector = (saaty_preference_matrix / saaty_preference_matrix.sum(axis=0)).mean(axis=1)
            sum_of_weights = (saaty_preference_matrix * priority_vector).sum(axis=1)    
            lambda_max = (sum_of_weights/priority_vector).mean()
            n = len(priority_vector)
            CI = (lambda_max - n)/(n -1) 
            CR = CI/SAATY_RI[n-1]
        if CR <= 0.1:
            self.weights = priority_vector.values.tolist()
            return {"consistency": True, "lambda": lambda_max, "CIndex": CI, "CRatio": CR}
        else:
            return {"consistency": False, "lambda": lambda_max, "CIndex": CI, "CRatio": CR}
                  
    def set_signals(self, signals):
        """
        Set the target function of criteria for maximization or minimization and save it into attribute: signals.

        e.g.
            set_signals([-1, 1, 1, 1, 1])

            or using contants MIN and MAX
            
            set_signals([MIN, MAX, MAX, MAX, MAX])

        :param signals: 1 or MAX for maximization or MIN, -1 for minimization.
        :type signals: list

        :returns: no value
        :rtype: none
        """
        if self.__check_signals(signals) is True:
            self.signals = signals
        else:
            raise ValueError("Invalid signals! It's must be a list of 1 or -1")

    def __set_normalization_method(self, normalization_method):
        if normalization_method in NORMALIZATION_METHODS:
            self.normalization_method = normalization_method 
        elif normalization_method is None:
            self.normalization_method == None
        else:
            raise ValueError("Invalid parameter! Use a method defined in constants. e.g. normalization_D, zScore etc...")

    def __normalize(self, ignore_signals=False):
        if self.normalization_method == Vector_:
            self.__norm_Vector(ignore_signals)
        elif self.normalization_method == LinearSum_:
            self.__norm_LinearSum(ignore_signals)
        elif self.normalization_method == LinearMinMax_:
            self.__norm_LinearMinMax(ignore_signals)
        elif self.normalization_method == LinearMax_:
            self.__norm_LinearMax(ignore_signals)
        elif self.normalization_method == EnhancedAccuracy_:
            self.__norm_EnhancedAccuracy(ignore_signals)
        elif self.normalization_method == Logarithmic_:
            self.__norm_Logarithmic(ignore_signals)
        else:
            self.df_normalized = self.df_original

    def __norm_ZScore(self, ignore_signals=False):
        normalized = (self.df_original.iloc[:, 1:] - self.df_original.iloc[:, 1:].mean(axis=0))/self.df_original.iloc[:, 1:].std(axis=0)
        self.df_normalized = pd.DataFrame(self.df_original.iloc[:, 0]).join(normalized)
               
        # make nomalized matrix based in signals
        normalized = pd.DataFrame(self.df_original.iloc[:, 0])
        if ignore_signals is True:
            normalized = normalized.join(normalized_max.iloc[:, :])
        else:
            i = 0
            for s in self.signals:
                if s == 1:
                    normalized = normalized.join(normalized_max.iloc[:, i])
                else:
                    normalized = normalized.join(normalized_min.iloc[:, i])
                i = i + 1
        
        self.df_normalized = normalized

    def __norm_EnhancedAccuracy(self, ignore_signals=False):
        # _max for maximizing 
        normalized_max = 1 - (self.df_original.iloc[:, 1:].max(axis=0)-self.df_original.iloc[:, 1:])/((self.df_original.iloc[:, 1:].max(axis=0)-self.df_original.iloc[:, 1:]).sum(axis=0))
        # _min for minimizing 
        normalized_min = 1 - (self.df_original.iloc[:, 1:]-self.df_original.iloc[:, 1:].min(axis=0))/((self.df_original.iloc[:, 1:]-self.df_original.iloc[:, 1:].min(axis=0)).sum(axis=0))
       
        # make nomalized matrix based in signals
        normalized = pd.DataFrame(self.df_original.iloc[:, 0])
        if ignore_signals is True:
            normalized = normalized.join(normalized_max.iloc[:, :])
        else:
            i = 0
            for s in self.signals:
                if s == 1:
                    normalized = normalized.join(normalized_max.iloc[:, i])
                else:
                    normalized = normalized.join(normalized_min.iloc[:, i])
                i = i + 1
        
        self.df_normalized = normalized

    def __norm_Logarithmic(self, ignore_signals=False):
        # _max for maximizing 
        normalized_max = np.log2(self.df_original.iloc[:, 1:]) / np.log2( np.log2(self.df_original.iloc[:, 1:]).prod(axis = 0) * self.df_original.iloc[:, 1:] )
        # _min for minimizing 
        normalized_min = 1 - ( (1 - normalized_max) / ( normalized_max.shape[0] - 1) )
          
        # make nomalized matrix based in signals
        normalized = pd.DataFrame(self.df_original.iloc[:, 0])
        if ignore_signals is True:
            normalized = normalized.join(normalized_max.iloc[:, :])
        else:
            i = 0
            for s in self.signals:
                if s == 1:
                    normalized = normalized.join(normalized_max.iloc[:, i])
                else:
                    normalized = normalized.join(normalized_min.iloc[:, i])
                i = i + 1
        
        self.df_normalized = normalized

    def __norm_LinearMax(self, ignore_signals=False):
        # _max for maximizing 
        normalized_max = (self.df_original.iloc[:, 1:]/self.df_original.iloc[:, 1:].max(axis=0))
        # _min for minimizing 
        normalized_min = (self.df_original.iloc[:, 1:].min(axis=0)/self.df_original.iloc[:, 1:])
               
        # make nomalized matrix based in signals
        normalized = pd.DataFrame(self.df_original.iloc[:, 0])
        if ignore_signals is True:
            normalized = normalized.join(normalized_max.iloc[:, :])
        else:
            i = 0
            for s in self.signals:
                if s == 1:
                    normalized = normalized.join(normalized_max.iloc[:, i])
                else:
                    normalized = normalized.join(normalized_min.iloc[:, i])
                i = i + 1
        
        self.df_normalized = normalized

    def __norm_LinearMinMax(self, ignore_signals=False):
        # _max for maximizing 
        normalized_max = (self.df_original.iloc[:, 1:]-self.df_original.iloc[:, 1:].min(axis=0))/(self.df_original.iloc[:, 1:].max(axis=0)-self.df_original.iloc[:, 1:].min(axis=0))
        # _min for minimizing 
        normalized_min = (self.df_original.iloc[:, 1:].max(axis=0)-self.df_original.iloc[:, 1:])/(self.df_original.iloc[:, 1:].max(axis=0)-self.df_original.iloc[:, 1:].min(axis=0))
               
        # make nomalized matrix based in signals
        normalized = pd.DataFrame(self.df_original.iloc[:, 0])
        if ignore_signals is True:
            normalized = normalized.join(normalized_max.iloc[:, :])
        else:
            i = 0
            for s in self.signals:
                if s == 1:
                    normalized = normalized.join(normalized_max.iloc[:, i])
                else:
                    normalized = normalized.join(normalized_min.iloc[:, i])
                i = i + 1
        
        self.df_normalized = normalized

    def __norm_LinearSum(self, ignore_signals=False):
        # _max for maximizing 
        normalized_max = self.df_original.iloc[:, 1:]/self.df_original.iloc[:, 1:].sum(axis=0)
        # _min for minimizing 
        normalized_min = (1/self.df_original.iloc[:, 1:])/(1/self.df_original.iloc[:, 1:]).sum(axis=0)
        
       
        # make nomalized matrix based in signals
        normalized = pd.DataFrame(self.df_original.iloc[:, 0])
        if ignore_signals is True:
            normalized = normalized.join(normalized_max.iloc[:, :])
        else:
            i = 0
            for s in self.signals:
                if s == 1:
                    normalized = normalized.join(normalized_max.iloc[:, i])
                else:
                    normalized = normalized.join(normalized_min.iloc[:, i])
                i = i + 1
        
        self.df_normalized = normalized

    def __norm_Vector(self, ignore_signals=False):
        # _max for maximizing 
        normalized_max = self.df_original.iloc[:, 1:]/np.sqrt(self.df_original.iloc[:, 1:].pow(2).sum(axis=0))
        # _min for minimizing 
        normalized_min = 1 - self.df_original.iloc[:, 1:]/np.sqrt(self.df_original.iloc[:, 1:].pow(2).sum(axis=0))
       
        # make nomalized matrix based in signals
        normalized = pd.DataFrame(self.df_original.iloc[:, 0])
        if ignore_signals is True:
            normalized = normalized.join(normalized_max.iloc[:, :])
        else:
            i = 0
            for s in self.signals:
                if s == 1:
                    normalized = normalized.join(normalized_max.iloc[:, i])
                else:
                    normalized = normalized.join(normalized_min.iloc[:, i])
                i = i + 1
        
        self.df_normalized = normalized


        # _max for maximizing 
        normalized_topsis = self.df_original.iloc[:, 1:]/np.sqrt(self.df_original.iloc[:, 1:].pow(2).sum(axis=0))
        
        # make nomalized matrix based in signals
        normalized = pd.DataFrame(self.df_original.iloc[:, 0])
        normalized = normalized.join(normalized_topsis.iloc[:, :])
        
        self.df_normalized = normalized

    def __weighting_from_normalized(self):
        weighted = self.df_normalized.iloc[:, 1:] * self.weights
        self.df_weighted = pd.DataFrame(self.df_original.iloc[:,0]).join(weighted)

    def __check_weights(self, weights, free=False):
        result = False
        if type(weights) == list:
            if len(self.df_original.columns.tolist()) - 1 == len(weights) \
               and ((all(isinstance(n, float) for n in weights) \
               and all((n >= 0 and n <= 1) for n in weights) \
               and round(sum(weights)) == 1) \
               or (free is True)):
                result = True
        return result

    def __check_signals(self, signals):
        result = False
        if type(signals) == list:
            if len(self.df_original.columns.tolist()) - 1 == len(signals) \
               and all(isinstance(n, int) for n in signals) \
               and all((n == 1 or n == -1) for n in signals):                
                result = True
        return result

    def __check_AHP_matrix(self, saaty_preference_matrix):
        if type(saaty_preference_matrix) == list:  
            saaty_preference_matrix = pd.DataFrame(saaty_preference_matrix)
            x, y = saaty_preference_matrix.shape 
            n = len(self.df_original.columns.values) -1
            if False in saaty_preference_matrix.iloc[:, :].isin(SAATY_SCALE).values or x != y or y != n:
                raise ValueError("Value not in Saaty Scale os wrong dimentions in matrix")
            else:
                for i in range(x):
                    for j in range(y):
                        if saaty_preference_matrix.iloc[i, j] == 1/saaty_preference_matrix.iloc[j, i] \
                           or saaty_preference_matrix.iloc[j, i] == 1/saaty_preference_matrix.iloc[i, j] \
                           or (saaty_preference_matrix.iloc[j, i] == 1 and i == j): 
                            pass
                        else:
                            raise ValueError("Incoherent values of preferences", saaty_preference_matrix.iloc[i, j], )
        return True

    def pretty_original(self, tablefmt='psql'):
        """
        Print a pretty table about specific attribute.

        :param tablefmt: (default=psql) format of output 
        :type: string constant: 'html', 'latex' or 'psql'

        :returns: print
        :rtype: string

        """
        return tabulate(self.df_original, headers='keys', tablefmt=tablefmt)

    def pretty_normalized(self, tablefmt='psql'):
        """
        Print a pretty table about specific attribute.

        :param tablefmt: (default=psql) format of output 
        :type: string constant: 'html', 'latex' or 'psql'

        :returns: print
        :rtype: string

        """
        return tabulate(self.df_normalized, headers='keys', tablefmt=tablefmt)

    def pretty_weighted(self, tablefmt='psql'):
        """
        Print a pretty table about specific attribute.

        :param tablefmt: (default=psql) format of output 
        :type: string constant: 'html', 'latex' or 'psql'

        :returns: print
        :rtype: string

        """        
        return tabulate(self.df_weighted, headers='keys', tablefmt=tablefmt)

    def pretty_decision(self, tablefmt='psql'):
        """
        Print a pretty table about specific attribute.

        :param tablefmt: (default=psql) format of output 
        :type: string constant: 'html', 'latex' or 'psql'

        :returns: print
        :rtype: string

        """
        return tabulate(self.df_decision, headers='keys', tablefmt=tablefmt)
        




from .mcda import MCDA, pd, np

class WPM(MCDA):
    def __init__(self):
        super().__init__()        
    
    def decide(self, normalization_method=None):  
        """
        Method for decision making using WPM. 

        :param normalization_method: (default=None) Must be defined when the problem data is different units,
        :type normalization_method: some value defined in NORMALIZATION_METHODS constant

        :returns: none value
        :rtype: none
        """   
        self._MCDA__set_normalization_method(normalization_method)
        self._MCDA__normalize()
        self._MCDA__weighting_from_normalized()
        self.__wpm()

    def __wpm(self):
        wpm = pd.DataFrame(self.df_weighted.iloc[:, 1:])
        wpm = pd.DataFrame(wpm.prod(axis=1), columns=["WPM"])

        i = np.arange(1, len(self.df_original.index)+1, 1)
        df_ranking = pd.DataFrame(self.df_original.iloc[:,0]).join(wpm).sort_values(by=["WPM"], ascending=False)
        df_ranking["rank"] = i

        self.df_decision = df_ranking.sort_index()




import setuptools

with open("READMEpypi.rst", "r") as fh:
    long_description = fh.read()


setuptools.setup(
      name="scikit-mcda",
      version="0.21.39",
      author="Antonio Horta",
      author_email='ajhorta@cybercrafter.com.br',
      description="Library for Multi-criteria Decision Aid Methods",
      long_description=long_description,
      long_description_content_type="text/markdown",
      url="https://gitlab.com/cybercrafter/scikit-mcda",
      install_requires=[
            "numpy~=1.18",
            "pandas~=1.0",
            "tabulate"
            ],
      license="Apache License 2.0",
      packages=setuptools.find_packages(),
      classifiers=[
        "Programming Language :: Python",
        "License :: OSI Approved :: Apache Software License",
        "Operating System :: OS Independent",
        "Topic :: Software Development :: Libraries",
        "Topic :: Scientific/Engineering :: Mathematics"
      ],
      python_requires='>=3.6',
      )
